import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.lang.Math;
import javax.swing.*;
import java.util.*;
import java.text.*;

public class Deadline extends Gravy {
    final long ONE_HOUR = 60 * 60 * 1000;
    final long ONE_DAY = 24 * ONE_HOUR;
    final long ONE_WEEK = 7 * ONE_DAY;
    final long ONE_MONTH = 4 * ONE_WEEK;

    /**
     * An inner class to collect the information to auto-scale the dial.
     */
    class TimeScale {
	/** Unit of each tick (from the finals above) */
	long tickUnit;
	/** Number of those units on the face */
	int numUnits;
	/** See Gravy - drawing parameter */
	int tickZero;
	/** See Gravy - drawing parameter */
	int numTicks;
	/** See Gravy - drawing parameter */
	int tickWidth;
	/** An extra label to give units */
	String label;
	/* Straightforward constructor. */
	public TimeScale(long u, int nu, int tz, int nt, int tw, String l) {
	    tickUnit = u;
	    numUnits = nu;
	    tickZero = tz;
	    numTicks = nt;
	    tickWidth = tw;
	    label = l;
	}
    }

    /** The deadline in question. */
    GregorianCalendar deadline;
    /** The various timescale information for auto-scaling */
    TimeScale[] scales;
    /** The current scale information */
    TimeScale scale;
    /** An array of DateFromats to try in order when parsing the deadline from
     * the parameters */
    DateFormat formats[];

    /**
     * Set scale to the most appropriate scaling parameters.  Scale is the
     * smallest scale that the current time will fit in.   The final element of
     * the label array is also changed to reflect new units, if any.
     * @return true if the scale has changed.
     */
    protected boolean setScale() {
	final long diff = 
	    deadline.getTime().getTime() - System.currentTimeMillis();
	final TimeScale oldScale = scale;
	
	for (int i = 0; i < scales.length; i++)
	    if ( diff < scales[i].tickUnit * scales[i].numUnits ) {
		scale = scales[i];
		break;
	    }
	// Make sure we pick a scale, especially for really distant events.
	if ( scale == null ) scale = scales[scales.length-1];
	if ( scale != oldScale ) {
	    label[label.length-1] = scale.label;
	    nticks = scale.numTicks;
	    tickZero = scale.tickZero;
	    tickWidth = scale.tickWidth;
	}
    
	return !(scale == oldScale);
    }

    /**
     * Return a GregorianCalendar representing the dateString.  dateString is
     * parsed using the various DateFormats in formats, and the first one that
     * can parse it is used.  If none can, a GregorianCalanedar from teh
     * default constructor is returned.
     * @param dateString the string to parse
     * @return the date as a GregorianCalendar
     */
    protected GregorianCalendar parseDate(String dateString) {

	for (int i = 0; i < formats.length; i++ ) {
	    try {
		Date d = formats[i].parse(dateString);
		GregorianCalendar cal = new GregorianCalendar();
		cal.setTime(d);
		return cal;
	    }
	    catch (ParseException e) { }
	}
	return new GregorianCalendar();
    }

    /**
     * Parse deadline and call the superclass readParameters routine.
     */
    void readParameters() {
	super.readParameters();

	try {
	    deadline = parseDate(getParameter("deadline"));
	}
	catch (Exception e) {
	    deadline = new GregorianCalendar();
	}
    }

    /**
     * Initialize the formats and autoscale info and then call the superclass
     * initializer.  Then add space for the scale label and set the scale.
     */
    public void init() {
	scales = new TimeScale[] {
	    new TimeScale(ONE_HOUR, 24, 24, 7, -4, "Hours"),
	    new TimeScale(ONE_DAY, 7, 7, 8, -1, "Days"),
	    new TimeScale(ONE_WEEK, 4, 4, 5, -1, "Weeks"),
	    new TimeScale(ONE_MONTH, 12, 12, 5, -3, "Months")
	};
	scale = null;
	formats = new DateFormat[] {
	    DateFormat.getDateTimeInstance(DateFormat.LONG,
		    DateFormat.LONG),
	    DateFormat.getDateTimeInstance(DateFormat.LONG,
		    DateFormat.MEDIUM),
	    DateFormat.getDateTimeInstance(DateFormat.MEDIUM,
		    DateFormat.MEDIUM),
	    DateFormat.getDateTimeInstance(DateFormat.SHORT,
		    DateFormat.MEDIUM),
	    DateFormat.getDateTimeInstance(DateFormat.LONG,
		    DateFormat.SHORT),
	    DateFormat.getDateTimeInstance(DateFormat.MEDIUM,
		    DateFormat.SHORT),
	    DateFormat.getDateTimeInstance(DateFormat.SHORT,
		    DateFormat.SHORT),
	    DateFormat.getDateInstance(DateFormat.LONG),
	    DateFormat.getDateInstance(DateFormat.MEDIUM),
	    DateFormat.getDateInstance(DateFormat.SHORT)
	};
	for (int i = 0; i < formats.length; i++ ) formats[i].setLenient(true);

	super.init();

	String [] nlabels = new String[label.length+2];
	System.arraycopy(label, 0, nlabels,0, label.length);
	nlabels[nlabels.length-2] = "";
	nlabels[nlabels.length-1] = null;
	label = nlabels;
	setScale();
	freq = 1000;
    }

    /**
     * Update the angle of the pointer.  Return true if it changed.  This may
     * rescale the dial.
     * @return true if the pointer moved
     */
    synchronized boolean updateAngle() {
	if ( previousPointerAngle != pointerAngle ) return false;
	
	final double amax = 3 * Math.PI/4;
	final double amin = -amax;
	boolean scaleChange = setScale();

	long diff = deadline.getTime().getTime() - 
	    System.currentTimeMillis();
	double units = (double) diff / scale.tickUnit;

	if ( units > 0.0 && units < scale.numUnits ) 
	    pointerAngle = 
		(amax - ( units / scale.numUnits ) * Math.abs(amax - amin));
	else 
	    pointerAngle = ( units < 0.0 ) ? amax : amin;

	// If the scale changed, force a redrawing of the whole dial once, and
	// the false return value prevents the double painting.
	if ( scaleChange ) {
	    mustDrawFace = true;
	    repaint();
	    return false;
	}
	return true;
    }

    /**
     * Simple command line invocation, for debugging.
     */
    static public void main(String argv[]) {
	JFrame f = new JFrame("Gravy");
	Deadline d = new Deadline();
	int x = 100, y=100;

	if ( argv.length > 0 ) x = Integer.parseInt(argv[0]);
	if ( argv.length > 1 ) y = Integer.parseInt(argv[1]);

	if ( x < 0 ) x = 100;
	if ( y < 0 ) y = 100;

	f.setVisible(true);
	// get the insets (f needs to be visible for that) and resize the frame
	// so that the dial has the requested area.
	Insets in = f.getInsets();
	f.setSize(x + in.left + in.right, y + in.top + in.bottom);
	f.setLocation(10,10);
	d.background = Color.white;
	d.init();
	if (argv.length > 2 ) d.deadline = d.parseDate(argv[2]);
	f.setContentPane(d);
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		System.exit(0);
	    }
	});
	f.validate();
	d.start();
    }
}
